/*
//  mod_pgsqllog.c
//
//
//  Author: Franco Gasperino
//  Date: 04/21/01
//
//  Revision: $Revision: 1.13 $
//  Last Modified: $Date: 2002/04/02 12:10:29 $
//  Last Committer: $Author: franco $
//
//  Copyright (C) 2001 Franco Gasperino
//
//  This program is free software; you can redistribute it and/or modify it under 
//  the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2, or (at your option) any later
//  version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along with
//  this program; if not, write to the Free Software Foundation, Inc., 675
//  Mass Ave, Cambridge, MA 02139, USA.
//
*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <assert.h>

#include "mod_pgsqllog.h"

/* configuration directive name-to-callback mapping */
static const command_rec cmd_table[] = {

    {"PgSqlLogHost", conf_host_action, NULL, OR_ALL, TAKE1,
     "PostgreSQL log host"},
    {"PgSqlLogPort", conf_port_action, NULL, OR_ALL, TAKE1,
     "PostgreSQL log port"},
    {"PgSqlLogDatabase", conf_database_action, NULL, OR_ALL, TAKE1,
     "PostgreSQL log database"},
    {"PgSqlLogUsername", conf_username_action, NULL, OR_ALL, TAKE1,
     "PostgreSQL log username"},
    {"PgSqlLogPassword", conf_password_action, NULL, OR_ALL, TAKE1,
     "PostgreSQL log password"},
    {"PgSqlLogTable", conf_table_action, NULL, OR_ALL, TAKE1,
     "PostgreSQL log table name"},
    {"PgSqlHandleCount", conf_handle_count_action, NULL, OR_ALL, TAKE1,
     "PostgreSQL connection pool count"},
    {"PgSqlConnectRetryCount", conf_connect_retry_count_action, NULL,
     OR_ALL, TAKE1,
     "PosgreSQL connection retry count"},
    NULL
};

/* module interface */
module MODULE_VAR_EXPORT pgsqllog_module = {
    STANDARD_MODULE_STUFF,
    init_handler,		/* module initializer                  */
    NULL,			/* create per-dir    config structures */
    NULL,			/* merge  per-dir    config structures */
    server_config_handler,	/* create per-server config structures */
    NULL,			/* merge  per-server config structures */
    cmd_table,			/* table of config file commands       */
    NULL,			/* [#8] MIME-typed-dispatched handlers */
    NULL,			/* [#1] URI to filename translation    */
    NULL,			/* [#4] validate user id from request  */
    NULL,			/* [#5] check if the user is ok _here_ */
    NULL,			/* [#3] check access by host address   */
    NULL,			/* [#6] determine MIME type            */
    NULL,			/* [#7] pre-run fixups                 */
    log_handler,		/* [#9] log a transaction              */
    NULL,			/* [#2] header parser                  */
    NULL,			/* child_init                          */
    NULL,			/* child_exit                          */
    NULL			/* [#0] post read-request              */
#ifdef EAPI
	, NULL,			/* EAPI: add_module                    */
    NULL,			/* EAPI: remove_module                 */
    NULL,			/* EAPI: rewrite_command               */
    NULL			/* EAPI: new_connection                */
#endif
};

/*
//  conf_host_action()
//
//  Callback invoked when configuration directive PgSqlLogHost
//  is parsed. Returns NULL on success, or error string on
//  failure.
//
//  parms   - address of cmd_parms struct
//  mconfig - dummy value
//  arg     - value parsed for directive key
*/

static const char *conf_host_action(cmd_parms * parms, void *mconfig,
				    const char *arg)
{
    sqlenv *env;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: conf_host_action() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(parms != NULL);

    /* don't accept NULL values */
    if (arg == NULL || strcmp(arg, NULL_STRING) == 0)
	return ("Empty or NULL string passed as PostgreSQL host name!");

    env = (sqlenv *) ap_get_module_config(parms->server->module_config,
					  &pgsqllog_module);

    /* store the value in the global config */
    if ((env->host = ap_pstrdup(parms->pool, arg)) == NULL)
	return (strerror(errno));

    return (NULL);
}

/*
//  conf_port_action()
//
//  Callback invoked when configuration directive PgSqlLogPort
//  is parsed. Returns NULL on success, or error string on
//  failure.
//
//  parms   - address of cmd_parms struct
//  mconfig - dummy value
//  arg     - value parsed for directive key
*/

static const char *conf_port_action(cmd_parms * parms, void *mconfig,
				    const char *arg)
{
    sqlenv *env;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: conf_port_action() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(parms != NULL);

    /* dont accept NULL values */
    if (arg == NULL || strcmp(arg, NULL_STRING) == 0)
	return ("Empty or NULL string passed as PostgreSQL port number!");

    env = (sqlenv *) ap_get_module_config(parms->server->module_config,
					  &pgsqllog_module);

    /* store the value in the global config */
    if ((env->port = ap_pstrdup(parms->pool, arg)) == NULL)
	return (strerror(errno));

    return (NULL);
}

/*
//  conf_database_action()
//
//  Callback invoked when configuration directive PgSqlLogDatabase
//  is parsed. Returns NULL on success, or error string on
//  failure.
//
//  parms   - address of cmd_parms struct
//  mconfig - dummy value
//  arg     - value parsed for directive key
*/

static const char *conf_database_action(cmd_parms * parms, void *mconfig,
					const char *arg)
{
    sqlenv *env;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: conf_database_action() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(parms != NULL);

    /* dont accept NULL values */
    if (arg == NULL || strcmp(arg, NULL_STRING) == 0)
	return
	    ("Empty or NULL string passed as PostgreSQL database name!");

    env = (sqlenv *) ap_get_module_config(parms->server->module_config,
					  &pgsqllog_module);

    /* store the value in the global config */
    if ((env->database = ap_pstrdup(parms->pool, arg)) == NULL)
	return (strerror(errno));

    return (NULL);
}

/*
//  conf_username_action()
//
//  Callback invoked when configuration directive PgSqlLogUsername
//  is parsed. Returns NULL on success, or error string on
//  failure.
//
//  parms   - address of cmd_parms struct
//  mconfig - dummy value
//  arg     - value parsed for directive key
*/

static const char *conf_username_action(cmd_parms * parms, void *mconfig,
					const char *arg)
{
    sqlenv *env;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: conf_username_action() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(parms != NULL);

    /* dont accept NULL values */
    if (arg == NULL || strcmp(arg, NULL_STRING) == 0)
	return ("Empty or NULL string passed as PostgreSQL username!");

    env = (sqlenv *) ap_get_module_config(parms->server->module_config,
					  &pgsqllog_module);

    /* store the value in the global config */
    if ((env->username = ap_pstrdup(parms->pool, arg)) == NULL)
	return (strerror(errno));

    return (NULL);
}

/*
//  conf_password_action()
//
//  Callback invoked when configuration directive PgSqlLogPassword
//  is parsed. Returns NULL on success, or error string on
//  failure.
//
//  parms   - address of cmd_parms struct
//  mconfig - dummy value
//  arg     - value parsed for directive key
*/

static const char *conf_password_action(cmd_parms * parms, void *mconfig,
					const char *arg)
{
    sqlenv *env;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: conf_password_action() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(parms != NULL);

    /* dont accept NULL values */
    if (arg == NULL || strcmp(arg, NULL_STRING) == 0)
	return ("Empty or NULL string passed as PostgreSQL password!");

    env = (sqlenv *) ap_get_module_config(parms->server->module_config,
					  &pgsqllog_module);

    /* store the value in the global config */
    if ((env->password = ap_pstrdup(parms->pool, arg)) == NULL)
	return (strerror(errno));

    return (NULL);
}

/*
//  conf_table_action()
//
//  Callback invoked when configuration directive PgSqlLogTable
//  is parsed. Returns NULL on success, or error string on
//  failure.
//
//  parms   - address of cmd_parms struct
//  mconfig - dummy value
//  arg     - value parsed for directive key
*/

static const char *conf_table_action(cmd_parms * parms, void *mconfig,
				     const char *arg)
{
    sqlenv *env;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: conf_table_action() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(parms != NULL);

    /* dont accept NULL values */
    if (arg == NULL || strcmp(arg, NULL_STRING) == 0)
	return ("Empty or NULL string passed as PostgreSQL log table!");

    env = (sqlenv *) ap_get_module_config(parms->server->module_config,
					  &pgsqllog_module);

    /* store the value in the global config */
    if ((env->table = ap_pstrdup(parms->pool, arg)) == NULL)
	return (strerror(errno));

    return (NULL);
}

/*
//  conf_handle_count_action()
//
//  Callback invoked when configuration directive PgSqlHandleCount
//  is parsed. Returns NULL on success, or error string on
//  failure.
//
//  parms   - address of cmd_parms struct
//  mconfig - dummy value
//  arg     - value parsed for directive key
*/

static const char *conf_handle_count_action(cmd_parms * parms,
					    void *mconfig, const char *arg)
{
    char *invalid;
    sqlenv *env;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: conf_handle_count_action() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(parms != NULL);

    /* dont accept NULL values */
    if (arg == NULL || strcmp(arg, NULL_STRING) == 0)
	return
	    ("Empty or NULL string passed as PostgreSQL handle pool count!");

    env = (sqlenv *) ap_get_module_config(parms->server->module_config,
					  &pgsqllog_module);

    /* convert the string to 8bit int */
    env->handle_count = (unsigned short) strtol(arg, &invalid, 10);
    if (env->handle_count == 0 || env->handle_count == USHRT_MAX
	|| (invalid != NULL && *invalid != '\0'))
	return ("Invalid value parsed for PgSqlHandleCount!");

    return (NULL);
}

static const char *conf_connect_retry_count_action(cmd_parms * parms,
						   void *mconfig,
						   const char *arg)
{
    char *invalid;
    sqlenv *env;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: conf_connect_retry_count_action() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(parms != NULL);

    /* dont accept NULL values */
    if (arg == NULL || strcmp(arg, NULL_STRING) == 0)
	return
	    ("Empty or NULL string passed as PostgreSQL connection retry count!");

    env = (sqlenv *) ap_get_module_config(parms->server->module_config,
					  &pgsqllog_module);

    /* convert the string to 8bit int */
    env->retry_count = (short) strtol(arg, &invalid, 10);
    if (env->retry_count == SHRT_MIN || env->retry_count == SHRT_MAX
	|| (invalid != NULL && *invalid != '\0'))
	return ("Invalid value parsed for PgSqlConnectRetryCount!");

    return (NULL);
}

/*
//  module_cleanup()
//
//  Callback invoked upon module exit, usually through normal server
//  shutdown methods. The connection pool is closed, and the IPC 
//  semaphores are destroyed.
//
//  data - callback data as specified in ap_register_cleanup()
*/

void module_cleanup(void *data)
{
    sqlenv *env;
    server_rec *s;
    unsigned short loop;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: module_cleanup() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(data != NULL);

    /* convience typecast */
    s = (server_rec *) data;

    env =
	(sqlenv *) ap_get_module_config(s->module_config,
					&pgsqllog_module);

    assert(env->handles != NULL);
    assert(env->handle_count > 0);

    for (loop = 0; loop < env->handle_count; loop++) {

	/* disconnect from the database */
	if (sql_disconnect(&env->handles[loop]) == -1)
	    fprintf(stderr,
		    "%s ERROR: Unable to close existing PostgreSQL logging connectivity!\n",
		    MODULE_NAME);

	/* destroy existing IPC semaphore */
	if (lock_destroy(&env->handles[loop]) == -1)
	    exit(1);
    }
}

/*
//  child_cleanup()
//
//  Callback invoked upon child process shutdown. No current
//  cleanup handling is required.
//
//  data - callback data as specified in ap_register_cleanup()
*/

void child_cleanup(void *data)
{

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: child_cleanup() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(data != NULL);

}

/*
//  init_handler()
//
//  Callback invoked upon server startup, after configuration directives
//  are parsed and saved. A persistant connection to the pgsql backend
//  is created here, which is used in all children.
//
//  s - address of server_rec struct
//  p - address of global pool
*/

void init_handler(server_rec * s, pool * p)
{
    unsigned short loop;
    sqlenv *env;

#ifdef _TRACE_C_CALLS
    fprintf(stderr, "%s DEBUG: init_handler() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(s != NULL && p != NULL);

    env =
	(sqlenv *) ap_get_module_config(s->module_config,
					&pgsqllog_module);

    /* if configuration directives arent set, dont attempt a connection */
    if (env->host == NULL || env->port == NULL || env->database == NULL
	|| env->username == NULL || env->password == NULL
	|| env->table == NULL) {
	fprintf(stderr,
		"%s ERROR: One or more required configuration directives not set...ignoring all future log calls!\n",
		MODULE_NAME);
	return;
    }

    /* create the pool of persistant connections */
    if ((env->handles =
	 (sqlhandle *) ap_palloc(p,
				 env->handle_count * sizeof(sqlhandle))) ==
	NULL) {
	fprintf(stderr,
		"%s ERROR: Failed to allocate apache pool memory: %s\n",
		MODULE_NAME, strerror(errno));
	return;
    }

    memset(env->handles, 0, env->handle_count * (sizeof(sqlhandle)));

    /* register cleanup functions */
    ap_register_cleanup(p, s, module_cleanup, child_cleanup);

    for (loop = 0; loop < env->handle_count; loop++) {

	/* connect to the database */
	if (sql_connect(env, &env->handles[loop], env->retry_count) == -1)
	    fprintf(stderr,
		    "%s ERROR: Unable to establish PostgreSQL logging connectivity!\n",
		    MODULE_NAME);

	/* create the IPC semaphore */
	if (lock_create(&env->handles[loop], loop) == -1)
	    exit(1);
    }
}

/*
//  server_config_handler()
//
//  Callback invoked upon server startup, before configuration
//  directive values are parsed. Initial preparation for server-wide
//  storage takes place here. Returns address of global configuration
//  storage data.
//
//  p - address of global pool
//  s - address of server_rec struct
*/

void *server_config_handler(pool * p, server_rec * s)
{
    sqlenv *env;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: server_config_handler() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(p != NULL && s != NULL);

    /* create the SQL environment, reset contents */
    if ((env = ap_palloc(p, sizeof(sqlenv))) == NULL) {
	fprintf(stderr, "%s ERROR: %s\n", MODULE_NAME, strerror(errno));
	exit(1);
    }

    memset(env, 0, sizeof(*env));

    env->handle_count = DEFAULT_HANDLE_COUNT;

    return ((void *) env);
}

/*
//  log_handler()
//
//  Callback invoked upon child process "logging" phase. All actions
//  are child process-specific. Returns OK upon successful logging,
//  or DECLINED upon failure (allowing for fall-through logging to
//  complete).
//
//  r - address of request_rec struct
*/

int log_handler(request_rec * r)
{
    sqlenv *env;
    int result;

#ifdef _TRACE_C_CALLS
    fprintf(stderr, "%s DEBUG: log_handler() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(r != NULL);

    env = (sqlenv *) ap_get_module_config(r->server->module_config,
					  &pgsqllog_module);

    /* don't log if configuration directives aren't set, nor if the pool is invalid */
    if (env->host == NULL || env->port == NULL || env->database == NULL
	|| env->username == NULL || env->password == NULL
	|| env->table == NULL || env->handles == NULL)
	return (DECLINED);

    assert(env->handles != NULL);

    result = sql_log(r, env);

    if (result > 0)
	return (OK);

    return (DECLINED);
}

/*
//  sql_connect()
//
//  Internal function to create a connection to the postgresql backend,
//  after configuration directives have been parsed. This connection
//  lasts until server shutdown, and is shared between all child
//  processes. If the connection is unsuccessful, reconnection attempts
//  are made until success or the retry value is exceeded.
//  Returns 1 upon success, or -1 upon failure.
//
//  env    - address of sqlenv struct
//  handle - address of sqlhandle struct
//  retry  - number of reconnection attempts to make (-1 = unlimited, 0 = none)
*/

int sql_connect(sqlenv * env, sqlhandle * handle, int retry)
{
    int count = 0;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: sql_connect() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(env != NULL && handle != NULL);

    /* dont use any options or tty values for the connection handle */
    if ((handle->connection =
	 PQsetdbLogin(env->host, env->port, NULL, NULL, env->database,
		      env->username, env->password)) == NULL) {
	fprintf(stderr,
		"%s ERROR: Failed to allocate connection handle from PostgreSQL backend!\n",
		MODULE_NAME);
	return (-1);
    }

    /* check the connection status */
    if (PQstatus(handle->connection) != CONNECTION_OK) {
	sql_error(handle, NULL);

	if (retry != 0) {

	    /* bad connection, try to reconnect */
	    while (count < retry) {
		sleep(1);

		fprintf(stderr, "%s INFO: Connection attempt %i..\n", MODULE_NAME, count);

		PQreset(handle->connection);

		if (PQstatus(handle->connection) == CONNECTION_OK)
		    break;
		else
		    count++;
	    }

	    if (count == retry) {
		fprintf(stderr,
			"%s ERROR: Unable to connect to PosgreSQL backend, attempted %i connections!\n",
			MODULE_NAME, count);
	    } else {
		handle->connected = 1;
		return (1);
	    }
	}

	/* reconnection attempts unsuccessful, bail out */
	fprintf(stderr, "%s ERROR: Unable to establish connection with PostgreSQL backend!\n", MODULE_NAME);

	PQfinish(handle->connection);
	handle->connection = NULL;

	return (-1);
    }

    handle->connected = 1;

    return (1);
}

/*
//  sql_disconnect()
//
//  Internal function to disconnect from the postgresql backend.
//  Currently unused.
//
//  handle - address of sqlhandle struct
*/

int sql_disconnect(sqlhandle * handle)
{
#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: sql_disconnect() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(handle != NULL);

    if (handle->connected != 0) {
	if (handle->connection != NULL) {
	    PQfinish(handle->connection);
	    handle->connection = NULL;
	}

	handle->connected = 0;
    }

    assert(handle->connection == NULL && handle->connected == 0);

    return (1);
}

/*
//  sql_log()
//
//  Internal function to perform post-request logging to the
//  postgresql database. The record insert is performed using
//  the persistent connection handle. Returns 1 on success,
//  or -1 upon error.
//
//  r   - address of request_rec struct
//  env - address of sqlenv struct
//
//  Contributors:
//    status & method logging - Brian Holman (me@brianholman.com)
*/

int sql_log(request_rec * r, sqlenv * env)
{
    char statement[MAX_STATEMENT_LEN];
    unsigned int loop, found;
    const char *value;
    PGresult *result;
    sqlhandle *handle;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: sql_log() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(r != NULL && env != NULL);
    assert(env->handles != NULL);

    while (1) {
	/* locate a connection in the pool to use */
	for (loop = 0, found = 0, handle = NULL;
	     loop < env->handle_count; loop++) {
	    /* attempt to grab the lock */
	    switch (lock_lock(&env->handles[loop])) {
	    case 1:		/* free handle found */
		handle = &env->handles[loop];
		found = 1;
		break;

	    case 0:		/* handle locked */
		continue;

	    case -1:
		return (-1);
	    }
	}

	/* no free sql handles, sleep for a bit */
	if (found == 0) {
	    usleep(HANDLE_WAIT_TIME);
	    continue;
	}

	/* check the connection status */
	if (PQstatus(handle->connection) != CONNECTION_OK) {
	    sql_error(handle, NULL);

	    fprintf(stderr, "%s ERROR: Attempting to recover connection to PostgreSQL backend..\n", MODULE_NAME);

	    /* attempt to reconnect */
	    sql_disconnect(handle);

	    if (sql_connect(env, handle, env->retry_count) == -1) {
		lock_unlock(handle);
		return (-1);
	    }

	    fprintf(stderr, "%s INFO: Connection recovered successfully\n", MODULE_NAME);
	}

	assert(handle != NULL);

	/* the worst we could do is truncate the string */
	snprintf(statement, sizeof(statement),
		 "INSERT INTO %s VALUES ('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%i', '%s', '%s', '%s', '%s', '%s', '%s', '%i', '%i', '%s')",
		 env->table, (r->hostname ? r->hostname : ""),
		 (r->protocol ? r->protocol : ""),
		 (r->content_type ? r->content_type : ""),
		 (r->content_encoding ? r->content_encoding : ""),
		 (r->uri ? r->uri : ""), (r->filename ? r->filename : ""),
		 (r->server->server_hostname ? r->server->
		  server_hostname : ""),
		 (r->server->port ? r->server->port : -1),
		 ((value =
		   ap_table_get(r->headers_in, "Referer")) ? value : ""),
		 ((value =
		   ap_table_get(r->headers_in,
				"User-Agent")) ? value : ""),
		 (r->connection->remote_ip ? r->connection->
		  remote_ip : ""),
		 (r->connection->remote_host ? r->connection->
		  remote_host : ""),
		 (r->connection->user ? r->connection->user : ""),
		 (r->
		  request_time ? epoch_timestamp(r->
						 request_time) : "NULL"),
		 (r->bytes_sent ? r->bytes_sent : -1),
		 (r->status ? r->status : -1),
		 (r->method ? r->method : ""));

	/* send the statement to the backend */
	if ((result = PQexec(handle->connection, statement)) == NULL) {
	    sql_error(handle, NULL);
	    lock_unlock(handle);
	    return (-1);
	}

	if (PQresultStatus(result) != PGRES_COMMAND_OK) {
	    sql_error(handle, result);
	    PQclear(result);
	    lock_unlock(handle);
	    return (-1);
	}

	PQclear(result);

	/* release the lock */
	if (lock_unlock(handle) == -1)
	    return (-1);

	return (1);
    }

    return (1);
}

/*
//  sql_error()
//
//  Internal function used to generate legible error messages, and
//  send them to STDERR.
//
//  handle - address of sqlhandle struct
//  result - address of postgresql result object
*/

void sql_error(sqlhandle * handle, PGresult * result)
{
#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: sql_error() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(handle != NULL);

    if (result == NULL)
	fprintf(stderr, "%s ERROR: (PostgreSQL) %s\n", MODULE_NAME,
		PQerrorMessage(handle->connection));
    else
	fprintf(stderr, "%s ERROR: (PostgreSQL) %s\n", MODULE_NAME,
		PQresultErrorMessage(result));

    fflush(stderr);
}

/*
//  lock_create()
//
//  Internal function used to create the IPC semaphore used to
//  lock the postgresql connection handle. Returns 1 on success,
//  or -1 on error.
//
//  handle - address of sqlhandle struct
*/

int lock_create(sqlhandle * handle, unsigned short keyval)
{
    key_t key;
    union semun semunion;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: lock_create() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(handle != NULL);
    assert(handle->semid == 0);

    if ((key = ftok("/", keyval)) == -1) {
	fprintf(stderr, "%s ERROR: Failed to create IPC key: %s\n",
		MODULE_NAME, strerror(errno));
	return (-1);
    }

    if ((handle->semid =
	 semget(key, 1, 0666 | IPC_CREAT | IPC_EXCL)) == -1) {

	if (errno == EEXIST) {

	    if ((handle->semid = semget(key, 1, 0)) == -1) {
		fprintf(stderr,
			"%s ERROR: Failed to obtain existing IPC key: %s\n",
			MODULE_NAME, strerror(errno));
		return (-1);
	    }

	    if (lock_destroy(handle) == -1)
		return (-1);
	    else
		return (lock_create(handle, keyval));
	}

	fprintf(stderr, "%s ERROR: Failed to create IPC semaphore: %s\n",
		MODULE_NAME, strerror(errno));

	return (-1);
    }

    semunion.val = 1;
    if (semctl(handle->semid, 0, SETVAL, semunion) == -1) {
	fprintf(stderr, "%s ERROR: Failed to reset IPC semaphore: %s\n",
		MODULE_NAME, strerror(errno));
	lock_destroy(handle);
	return (-1);
    }

    return (1);
}

/*
//  lock_destroy()
//
//  Internal function used to destroy the IPC semaphore used to
//  lock the postgresql connection handle. Returns 1 on success,
//  or -1 on error.
//
//  handle - address of sqlhandle struct
*/

int lock_destroy(sqlhandle * handle)
{
    union semun semunion;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: lock_destroy() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(handle != NULL);
    assert(handle->semid != 0);

    if (semctl(handle->semid, 0, IPC_RMID, semunion) == -1) {
	fprintf(stderr, "%s ERROR: Failed to destroy IPC semaphore: %s\n",
		MODULE_NAME, strerror(errno));
	return (-1);
    }

    handle->semid = 0;

    return (1);
}

/*
//  lock_lock()
//
//  Internal function providing a wrapper for acquiring
//  the IPC semaphore lock on the postgresql connection
//  handle. Returns 1 on success, 0 on locked, or -1 on error.
//
//  handle - address of sqlhandle struct
*/

int lock_lock(sqlhandle * handle)
{
    struct sembuf sembuffer;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: lock_lock() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(handle != NULL);
    assert(handle->semid != 0);

    sembuffer.sem_num = 0;
    sembuffer.sem_op = -1;
    sembuffer.sem_flg = SEM_UNDO | IPC_NOWAIT;

    if (semop(handle->semid, &sembuffer, 1) == -1) {

	/* lock held by other proc */
	if (errno == EAGAIN)
	    return (0);

	fprintf(stderr,
		"%s ERROR: Failed to acquire IPC semaphore: %s - module may deadlock!!\n",
		MODULE_NAME, strerror(errno));
	return (-1);
    }

    return (1);
}

/*
//  lock_unlock()
//
//  Internal function providing a wrapper for releasing
//  the IPC semaphore lock on the postgresql connection handle.
//  Returns 1 on success, or -1 on error.
//
//  handle - address of sqlhandle struct
*/

int lock_unlock(sqlhandle * handle)
{
    struct sembuf sembuffer;

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: lock_unlock() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    assert(handle != NULL);
    assert(handle->semid != 0);

    sembuffer.sem_num = 0;
    sembuffer.sem_op = 1;
    sembuffer.sem_flg = SEM_UNDO;

    if (semop(handle->semid, &sembuffer, 1) == -1) {
	fprintf(stderr,
		"%s ERROR: Failed to release IPC semaphore: %s - module may deadlock!!\n",
		MODULE_NAME, strerror(errno));
	return (-1);
    }

    return (1);
}

/*
//  epoch_timestamp()
//
//  Convert a integer timestamp value (epoch) to readable timestamp.
//  Format used is yy/mm/dd hh/mm/ss UTC offset, which may need to be
//  modified for specific PostgreSQL setups. Returns timestamp string.
//
//  timer - epoch integer value to convert
//
//  Contributed by Brian Holman (me@brianholman.com)
*/

const char *epoch_timestamp(time_t timer)
{
    struct tm *tblock;
    static char tmstr[MAX_TIMESTAMP_LEN];

#ifdef _TRACE_C_CALLS
    fprintf(stderr,
	    "%s DEBUG: epoch_timestamp() (file: %s, line: %i)\n",
	    MODULE_NAME, __FILE__, __LINE__);
#endif

    if ((tblock = localtime(&timer)) == NULL) {
	fprintf(stderr,
		"%s ERROR: Failed to convert value %i to tm struct: localtime()\n",
		timer);
	return ("NULL");
    }

    if (strftime(tmstr, MAX_TIMESTAMP_LEN, "%G-%m-%d %H:%M:%S %Z", tblock)
	== 0) {
	fprintf(stderr,
		"%s ERROR: Failed to convert value %i into timestamp: strftime()\n",
		timer);
	return ("NULL");
    }

    return (tmstr);
}
